"""Tests of argument validation"""


import pytest

from hipscat_import.catalog.arguments import ImportArguments, check_healpix_order_range

# pylint: disable=protected-access


def test_none():
    """No arguments provided. Should error for required args."""
    with pytest.raises(ValueError):
        ImportArguments()


def test_empty_required(small_sky_parts_dir, tmp_path):
    """*Most* required arguments are provided."""
    ## Input format is missing
    with pytest.raises(ValueError, match="input_format"):
        ImportArguments(
            output_artifact_name="catalog",
            input_format=None,
            input_path=small_sky_parts_dir,
            output_path=tmp_path,
        )

    ## Input path is missing
    with pytest.raises(ValueError, match="input_file"):
        ImportArguments(
            output_artifact_name="catalog",
            input_path="",
            input_format="csv",
            output_path=tmp_path,
            overwrite=True,
        )


def test_invalid_paths(blank_data_dir, tmp_path):
    """Required arguments are provided, but paths aren't found."""
    ## Prove that it works with required args
    ImportArguments(
        output_artifact_name="catalog",
        input_path=blank_data_dir,
        output_path=tmp_path,
        input_format="csv",
        progress_bar=False,
    )

    ## Bad input path
    with pytest.raises(FileNotFoundError):
        ImportArguments(
            output_artifact_name="catalog",
            input_path="path",
            output_path=tmp_path,
            overwrite=True,
            input_format="csv",
        )

    ## Input path has no files
    with pytest.raises(FileNotFoundError):
        ImportArguments(
            output_artifact_name="catalog",
            input_path=blank_data_dir,
            output_path=tmp_path,
            overwrite=True,
            input_format="parquet",
        )


def test_good_paths(blank_data_dir, blank_data_file, tmp_path):
    """Required arguments are provided, and paths are found."""
    tmp_path_str = str(tmp_path)
    args = ImportArguments(
        output_artifact_name="catalog",
        input_path=blank_data_dir,
        input_format="csv",
        output_path=tmp_path_str,
        tmp_dir=tmp_path_str,
        progress_bar=False,
    )
    assert args.input_path == blank_data_dir
    assert len(args.input_paths) == 1
    assert blank_data_file in args.input_paths[0]


def test_multiple_files_in_path(small_sky_parts_dir, tmp_path):
    """Required arguments are provided, and paths are found."""
    args = ImportArguments(
        output_artifact_name="catalog",
        input_path=small_sky_parts_dir,
        input_format="csv",
        output_path=tmp_path,
        progress_bar=False,
    )
    assert args.input_path == small_sky_parts_dir
    assert len(args.input_paths) == 5


def test_single_debug_file(formats_headers_csv, tmp_path):
    """Required arguments are provided, and paths are found."""
    args = ImportArguments(
        output_artifact_name="catalog",
        input_file_list=[formats_headers_csv],
        input_format="csv",
        output_path=tmp_path,
        progress_bar=False,
    )
    assert len(args.input_paths) == 1
    assert args.input_paths[0] == formats_headers_csv


def test_healpix_args(blank_data_dir, tmp_path):
    """Test errors for healpix partitioning arguments"""
    with pytest.raises(ValueError, match="highest_healpix_order"):
        ImportArguments(
            output_artifact_name="catalog",
            input_path=blank_data_dir,
            input_format="csv",
            output_path=tmp_path,
            highest_healpix_order=30,
            overwrite=True,
        )
    with pytest.raises(ValueError, match="pixel_threshold"):
        ImportArguments(
            output_artifact_name="catalog",
            input_path=blank_data_dir,
            input_format="csv",
            output_path=tmp_path,
            pixel_threshold=3,
            overwrite=True,
        )
    with pytest.raises(ValueError, match="constant_healpix_order"):
        ImportArguments(
            output_artifact_name="catalog",
            input_path=blank_data_dir,
            input_format="csv",
            output_path=tmp_path,
            constant_healpix_order=30,
            overwrite=True,
        )


def test_catalog_type(blank_data_dir, tmp_path):
    """Test errors for catalog types."""
    with pytest.raises(ValueError, match="catalog_type"):
        ImportArguments(
            output_artifact_name="catalog",
            catalog_type=None,
            input_path=blank_data_dir,
            input_format="csv",
            output_path=tmp_path,
        )

    with pytest.raises(ValueError, match="catalog_type"):
        ImportArguments(
            output_artifact_name="catalog",
            catalog_type="association",
            input_path=blank_data_dir,
            input_format="csv",
            output_path=tmp_path,
        )


def test_to_catalog_info(blank_data_dir, tmp_path):
    """Verify creation of catalog parameters for catalog to be created."""
    args = ImportArguments(
        output_artifact_name="catalog",
        input_path=blank_data_dir,
        input_format="csv",
        output_path=tmp_path,
        tmp_dir=tmp_path,
        progress_bar=False,
    )
    catalog_info = args.to_catalog_info(total_rows=10)
    assert catalog_info.catalog_name == "catalog"
    assert catalog_info.total_rows == 10


def test_provenance_info(blank_data_dir, tmp_path):
    """Verify that provenance info includes catalog-specific fields."""
    args = ImportArguments(
        output_artifact_name="catalog",
        input_path=blank_data_dir,
        input_format="csv",
        output_path=tmp_path,
        tmp_dir=tmp_path,
        progress_bar=False,
    )

    runtime_args = args.provenance_info()["runtime_args"]
    assert "epoch" in runtime_args


def test_check_healpix_order_range():
    """Test method check_healpix_order_range"""
    check_healpix_order_range(5, "order_field")
    check_healpix_order_range(5, "order_field", lower_bound=0, upper_bound=19)

    with pytest.raises(ValueError, match="positive"):
        check_healpix_order_range(5, "order_field", lower_bound=-1)

    with pytest.raises(ValueError, match="19"):
        check_healpix_order_range(5, "order_field", upper_bound=20)

    with pytest.raises(ValueError, match="order_field"):
        check_healpix_order_range(-1, "order_field")
    with pytest.raises(ValueError, match="order_field"):
        check_healpix_order_range(30, "order_field")

    with pytest.raises(TypeError, match="not supported"):
        check_healpix_order_range("two", "order_field")
    with pytest.raises(TypeError, match="not supported"):
        check_healpix_order_range(5, "order_field", upper_bound="ten")



"""Flow control and pipeline entry points."""
import smtplib
from email.message import EmailMessage

from dask.distributed import Client

import hipscat_import.catalog.run_import as catalog_runner
import hipscat_import.index.run_index as index_runner
import hipscat_import.margin_cache.margin_cache as margin_runner
import hipscat_import.soap.run_soap as soap_runner
import hipscat_import.verification.run_verification as verification_runner
from hipscat_import.catalog.arguments import ImportArguments
from hipscat_import.index.arguments import IndexArguments
from hipscat_import.margin_cache.margin_cache_arguments import MarginCacheArguments
from hipscat_import.runtime_arguments import RuntimeArguments
from hipscat_import.soap.arguments import SoapArguments
from hipscat_import.verification.arguments import VerificationArguments

# pragma: no cover


def pipeline(args: RuntimeArguments):
    """Pipeline that creates its own client from the provided runtime arguments"""
    with Client(
        local_directory=args.dask_tmp,
        n_workers=args.dask_n_workers,
        threads_per_worker=args.dask_threads_per_worker,
    ) as client:
        pipeline_with_client(args, client)


def pipeline_with_client(args: RuntimeArguments, client: Client):
    """Pipeline that is run using an existing client.

    This can be useful in tests, or when a dask client requires some more complex
    configuraion.
    """
    try:
        if not args:
            raise ValueError("args is required and should be subclass of RuntimeArguments")

        if isinstance(args, ImportArguments):
            catalog_runner.run(args, client)
        elif isinstance(args, IndexArguments):
            index_runner.run(args)
        elif isinstance(args, MarginCacheArguments):
            margin_runner.generate_margin_cache(args, client)
        elif isinstance(args, SoapArguments):
            soap_runner.run(args, client)
        elif isinstance(args, VerificationArguments):
            verification_runner.run(args)
        else:
            raise ValueError("unknown args type")

        _send_success_email(args)
    except Exception as exception:  # pylint: disable=broad-exception-caught
        if args.completion_email_address:
            _send_failure_email(args, exception)
        raise exception


def _send_failure_email(args: RuntimeArguments, exception: Exception):
    message = EmailMessage()
    message["Subject"] = "hipscat-import failure."
    message["To"] = args.completion_email_address
    message.set_content(
        f"output_artifact_name: {args.output_artifact_name}"
        "\n\nSee logs for more details"
        f"\n\nFailed with message:\n\n{exception}"
    )

    _send_email(message)


def _send_success_email(args):
    if not args.completion_email_address:
        return
    message = EmailMessage()
    message["Subject"] = "hipscat-import success."
    message["To"] = args.completion_email_address
    message.set_content(f"output_artifact_name: {args.output_artifact_name}")

    _send_email(message)


def _send_email(message: EmailMessage):
    message["From"] = "updates@lsdb.io"

    # Send the message via our own SMTP server.
    with smtplib.SMTP("localhost") as server:
        server.send_message(message)



"""Test full execution of SOAP."""
import pytest
from hipscat.catalog.association_catalog.association_catalog import AssociationCatalog

import hipscat_import.soap.run_soap as runner
from hipscat_import.soap.arguments import SoapArguments


def test_empty_args():
    """Runner should fail with empty arguments"""
    with pytest.raises(TypeError):
        runner.run(None, None)


def test_bad_args():
    """Runner should fail with mis-typed arguments"""
    args = {"output_catalog_name": "bad_arg_type"}
    with pytest.raises(TypeError):
        runner.run(args, None)


@pytest.mark.dask
def test_object_to_source(dask_client, small_sky_soap_args):
    """Test creating association between object and source catalogs."""
    runner.run(small_sky_soap_args, dask_client)

    ## Check that the association data can be parsed as a valid association catalog.
    catalog = AssociationCatalog.read_from_hipscat(small_sky_soap_args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == small_sky_soap_args.catalog_path
    assert len(catalog.get_join_pixels()) == 14
    assert catalog.catalog_info.total_rows == 17161


@pytest.mark.dask
def test_object_to_source_with_leaves(
    dask_client, tmp_path, small_sky_object_catalog, small_sky_source_catalog
):
    """Test creating association between object and source catalogs."""
    small_sky_soap_args = SoapArguments(
        object_catalog_dir=small_sky_object_catalog,
        object_id_column="id",
        source_catalog_dir=small_sky_source_catalog,
        source_object_id_column="object_id",
        output_path=tmp_path,
        overwrite=True,
        progress_bar=False,
        write_leaf_files=True,
        source_id_column="source_id",
        output_artifact_name="small_sky_object_to_source",
    )
    runner.run(small_sky_soap_args, dask_client)

    ## Check that the association data can be parsed as a valid association catalog.
    catalog = AssociationCatalog.read_from_hipscat(small_sky_soap_args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == small_sky_soap_args.catalog_path
    assert len(catalog.get_join_pixels()) == 14
    assert catalog.catalog_info.total_rows == 17161



"""Utility to hold all arguments required throughout verification pipeline"""

from dataclasses import dataclass, field
from typing import List, Optional

from hipscat.catalog import Catalog
from hipscat.io.validation import is_valid_catalog

from hipscat_import.runtime_arguments import RuntimeArguments


@dataclass
class VerificationArguments(RuntimeArguments):
    """Data class for holding verification arguments"""

    ## Input
    input_catalog_path: str = ""
    """Path to an existing catalog that will be inspected."""
    input_catalog: Optional[Catalog] = None
    """In-memory representation of a catalog. If not provided, it will be loaded
    from the input_catalog_path."""

    ## Verification options
    field_distribution_cols: List[str] = field(default_factory=list)
    """List of fields to get the overall distribution for. e.g. ["ra", "dec"].
    Should be valid columns in the parquet files."""

    def __post_init__(self):
        self._check_arguments()

    def _check_arguments(self):
        super()._check_arguments()
        if not self.input_catalog_path and not self.input_catalog:
            raise ValueError("input catalog is required (either input_catalog_path or input_catalog)")
        if not self.input_catalog:
            if not is_valid_catalog(self.input_catalog_path):
                raise ValueError("input_catalog_path not a valid catalog")
            self.input_catalog = Catalog.read_from_hipscat(catalog_path=self.input_catalog_path)
        if not self.input_catalog_path:
            self.input_catalog_path = self.input_catalog.catalog_path

    def additional_runtime_provenance_info(self) -> dict:
        return {
            "pipeline": "verification pipeline",
            "input_catalog_path": str(self.input_catalog_path),
            "field_distribution_cols": self.field_distribution_cols,
        }



import pytest

import hipscat_import.verification.run_verification as runner
from hipscat_import.verification.arguments import VerificationArguments


def test_bad_args():
    """Runner should fail with empty or mis-typed arguments"""
    with pytest.raises(TypeError, match="VerificationArguments"):
        runner.run(None)

    args = {"output_artifact_name": "bad_arg_type"}
    with pytest.raises(TypeError, match="VerificationArguments"):
        runner.run(args)


def test_no_implementation(tmp_path, small_sky_object_catalog):
    """Womp womp. Test that we don't have a verification pipeline implemented"""
    args = VerificationArguments(
        input_catalog_path=small_sky_object_catalog,
        output_path=tmp_path,
        output_artifact_name="small_sky_object_verification_report",
    )
    with pytest.raises(NotImplementedError, match="not yet implemented"):
        runner.run(args)



"""Inner methods for SOAP"""

from typing import List

import numpy as np
import pandas as pd
import pyarrow.parquet as pq
from hipscat.catalog.association_catalog.partition_join_info import PartitionJoinInfo
from hipscat.io import FilePointer, file_io, paths
from hipscat.io.parquet_metadata import get_healpix_pixel_from_metadata
from hipscat.pixel_math.healpix_pixel import HealpixPixel
from hipscat.pixel_math.healpix_pixel_function import get_pixel_argsort

from hipscat_import.pipeline_resume_plan import get_pixel_cache_directory
from hipscat_import.soap.arguments import SoapArguments
from hipscat_import.soap.resume_plan import SoapPlan


def _count_joins_for_object(source_data, source_pixel, object_pixel, soap_args):
    object_path = paths.pixel_catalog_file(
        catalog_base_dir=soap_args.object_catalog_dir,
        pixel_order=object_pixel.order,
        pixel_number=object_pixel.pixel,
    )
    object_data = file_io.load_parquet_to_pandas(object_path, columns=[soap_args.object_id_column]).set_index(
        soap_args.object_id_column
    )

    joined_data = source_data.merge(object_data, how="inner", left_index=True, right_index=True)

    rows_written = len(joined_data)
    if not soap_args.write_leaf_files or rows_written == 0:
        return rows_written

    pixel_dir = get_pixel_cache_directory(soap_args.tmp_path, object_pixel)
    file_io.make_directory(pixel_dir, exist_ok=True)
    output_file = file_io.append_paths_to_pointer(
        pixel_dir, f"source_{source_pixel.order}_{source_pixel.pixel}.parquet"
    )
    joined_data = joined_data.reset_index()

    joined_data["Norder"] = np.full(rows_written, fill_value=object_pixel.order, dtype=np.uint8)
    joined_data["Dir"] = np.full(rows_written, fill_value=object_pixel.dir, dtype=np.uint64)
    joined_data["Npix"] = np.full(rows_written, fill_value=object_pixel.pixel, dtype=np.uint64)

    joined_data["join_Norder"] = np.full(rows_written, fill_value=source_pixel.order, dtype=np.uint8)
    joined_data["join_Dir"] = np.full(rows_written, fill_value=source_pixel.dir, dtype=np.uint64)
    joined_data["join_Npix"] = np.full(rows_written, fill_value=source_pixel.pixel, dtype=np.uint64)

    joined_data.to_parquet(output_file, index=True)

    return rows_written


def _write_count_results(cache_path, source_healpix, results):
    """Build a nice dataframe with pretty columns and rows"""
    num_results = len(results)
    dataframe = pd.DataFrame(results, columns=["Norder", "Npix", "num_rows"])

    dataframe["Dir"] = [int(order / 10_000) * 10_000 if order >= 0 else -1 for order, _, _ in results]
    dataframe["join_Norder"] = np.full(num_results, fill_value=source_healpix.order, dtype=np.uint8)
    dataframe["join_Dir"] = np.full(num_results, fill_value=source_healpix.dir, dtype=np.uint64)
    dataframe["join_Npix"] = np.full(num_results, fill_value=source_healpix.pixel, dtype=np.uint64)

    ## Reorder columns.
    dataframe = dataframe[["Norder", "Dir", "Npix", "join_Norder", "join_Dir", "join_Npix", "num_rows"]]

    file_io.write_dataframe_to_csv(
        dataframe=dataframe,
        file_pointer=file_io.append_paths_to_pointer(
            cache_path, f"{source_healpix.order}_{source_healpix.pixel}.csv"
        ),
        index=False,
    )


def count_joins(soap_args: SoapArguments, source_pixel: HealpixPixel, object_pixels: List[HealpixPixel]):
    """Count the number of equijoined sources in the object pixels.
    If any un-joined source pixels remain, stretch out to neighboring object pixels.

    Args:
        soap_args(`hipscat_import.soap.SoapArguments`): set of arguments for pipeline execution
        source_pixel(HealpixPixel): order and pixel for the source catalog single pixel.
        object_pixels(List[HealpixPixel]): set of tuples of order and pixel for the partitions
            of the object catalog to be joined.
    """
    source_path = paths.pixel_catalog_file(
        catalog_base_dir=file_io.get_file_pointer_from_path(soap_args.source_catalog_dir),
        pixel_order=source_pixel.order,
        pixel_number=source_pixel.pixel,
    )
    if soap_args.write_leaf_files:
        read_columns = [soap_args.source_object_id_column, soap_args.source_id_column]
    else:
        read_columns = [soap_args.source_object_id_column]
    source_data = file_io.load_parquet_to_pandas(source_path, columns=read_columns).set_index(
        soap_args.source_object_id_column
    )

    remaining_sources = len(source_data)
    results = []

    for object_pixel in object_pixels:
        if remaining_sources < 1:
            break
        join_count = _count_joins_for_object(
            source_data,
            source_pixel,
            object_pixel,
            soap_args,
        )
        results.append([object_pixel.order, object_pixel.pixel, join_count])
        remaining_sources -= join_count

    ## mark that some sources were not joined
    if remaining_sources > 0:
        results.append([-1, -1, remaining_sources])

    _write_count_results(soap_args.tmp_path, source_pixel, results)


def combine_partial_results(input_path, output_path) -> int:
    """Combine many partial CSVs into single partition join info.
    Also write out a debug file with counts of unmatched sources, if any.

    Args:
        input_path(str): intermediate directory with partial result CSVs. likely, the
            directory used in the previous `count_joins` call as `cache_path`
        output_path(str): directory to write the combined results CSVs.

    Returns:
        integer that is the sum of all matched num_rows.
    """
    partial_files = file_io.find_files_matching_path(input_path, "**.csv")
    partials = []

    for partial_file in partial_files:
        partials.append(file_io.load_csv_to_pandas(partial_file))

    dataframe = pd.concat(partials)

    matched = dataframe.loc[dataframe["Norder"] != -1]
    matched = matched.loc[matched["num_rows"] > 0]
    unmatched = dataframe.loc[dataframe["Norder"] == -1]

    file_io.write_dataframe_to_csv(
        dataframe=matched,
        file_pointer=file_io.append_paths_to_pointer(output_path, "partition_join_info.csv"),
        index=False,
    )

    if len(unmatched) > 0:
        file_io.write_dataframe_to_csv(
            dataframe=unmatched,
            file_pointer=file_io.append_paths_to_pointer(output_path, "unmatched_sources.csv"),
            index=False,
        )

    primary_only = matched.groupby(["Norder", "Dir", "Npix"])["num_rows"].sum().reset_index()
    file_io.write_dataframe_to_csv(
        dataframe=primary_only,
        file_pointer=file_io.append_paths_to_pointer(output_path, "partition_info.csv"),
        index=False,
    )

    join_info = PartitionJoinInfo(matched)
    join_info.write_to_metadata_files(output_path)

    return primary_only["num_rows"].sum()


def reduce_joins(
    soap_args: SoapArguments, object_pixel: HealpixPixel, object_key: str, delete_input_files: bool = True
):
    """Reduce join tables into one parquet file per object-pixel, with one row-group
    inside per source pixel."""
    pixel_dir = get_pixel_cache_directory(soap_args.tmp_path, object_pixel)
    # If there's no directory, this implies there were no matches to this object pixel
    # earlier in the pipeline. Move on.
    if not file_io.does_file_or_directory_exist(pixel_dir):
        return
    # Find all of the constituent files / source pixels. Create a list of PyArrow Tables from those
    # parquet files. We need to know the schema before we create the ParquetWriter.
    shard_file_list = file_io.find_files_matching_path(pixel_dir, "source**.parquet")

    if len(shard_file_list) == 0:
        return

    ## We want to order the row groups in a "breadth-first" sorting. Determine our sorting
    ## via the metadata, then read the tables in using that sorting.
    healpix_pixels = []
    for shard_file_name in shard_file_list:
        healpix_pixels.append(
            get_healpix_pixel_from_metadata(pq.read_metadata(shard_file_name), "join_Norder", "join_Npix")
        )

    argsort = get_pixel_argsort(healpix_pixels)
    shard_file_list = np.array(shard_file_list)[argsort]

    shards = []
    for shard_file_name in shard_file_list:
        shards.append(pq.read_table(shard_file_name))

    # Write all of the shards into a single parquet file, one row-group-per-shard.
    starting_catalog_path = FilePointer(str(soap_args.catalog_path))
    destination_dir = paths.pixel_directory(starting_catalog_path, object_pixel.order, object_pixel.pixel)
    file_io.make_directory(destination_dir, exist_ok=True)

    output_file = paths.pixel_catalog_file(starting_catalog_path, object_pixel.order, object_pixel.pixel)
    with pq.ParquetWriter(output_file, shards[0].schema) as writer:
        for table in shards:
            writer.write_table(table)

    # Delete the intermediate shards.
    if delete_input_files:
        file_io.remove_directory(pixel_dir, ignore_errors=True)

    SoapPlan.reducing_key_done(tmp_path=soap_args.tmp_path, reducing_key=object_key)



from dataclasses import dataclass

from hipscat.catalog.association_catalog.association_catalog import AssociationCatalogInfo
from hipscat.catalog.catalog_type import CatalogType
from hipscat.io.validation import is_valid_catalog

from hipscat_import.runtime_arguments import RuntimeArguments


@dataclass
class SoapArguments(RuntimeArguments):
    """Data class for holding source-object association arguments"""

    ## Input - Object catalog
    object_catalog_dir: str = ""
    object_id_column: str = ""

    ## Input - Source catalog
    source_catalog_dir: str = ""
    source_object_id_column: str = ""
    source_id_column: str = ""

    resume: bool = True
    """if there are existing intermediate resume files, should we
    read those and continue to run the pipeline where we left off"""
    write_leaf_files: bool = False
    """Should we also write out leaf parquet files (e.g. Norder/Dir/Npix.parquet)
    that represent the full association table"""

    compute_partition_size: int = 1_000_000_000

    def __post_init__(self):
        self._check_arguments()

    def _check_arguments(self):
        super()._check_arguments()
        if not self.object_catalog_dir:
            raise ValueError("object_catalog_dir is required")
        if not self.object_id_column:
            raise ValueError("object_id_column is required")
        if not is_valid_catalog(self.object_catalog_dir):
            raise ValueError("object_catalog_dir not a valid catalog")

        if not self.source_catalog_dir:
            raise ValueError("source_catalog_dir is required")
        if not self.source_object_id_column:
            raise ValueError("source_object_id_column is required")
        if not is_valid_catalog(self.source_catalog_dir):
            raise ValueError("source_catalog_dir not a valid catalog")

        if self.compute_partition_size < 100_000:
            raise ValueError("compute_partition_size must be at least 100_000")

    def to_catalog_info(self, total_rows) -> AssociationCatalogInfo:
        """Catalog-type-specific dataset info."""
        info = {
            "catalog_name": self.output_artifact_name,
            "catalog_type": CatalogType.ASSOCIATION,
            "total_rows": total_rows,
            "primary_column": self.object_id_column,
            "primary_catalog": str(self.object_catalog_dir),
            "join_column": self.source_object_id_column,
            "join_catalog": str(self.source_catalog_dir),
        }
        return AssociationCatalogInfo(**info)

    def additional_runtime_provenance_info(self) -> dict:
        return {
            "object_catalog_dir": self.object_catalog_dir,
            "object_id_column": self.object_id_column,
            "source_catalog_dir": self.source_catalog_dir,
            "source_object_id_column": self.source_object_id_column,
            "source_id_column": self.source_id_column,
            "compute_partition_size": self.compute_partition_size,
            "write_leaf_files": self.write_leaf_files,
        }



"""Import a set of non-hipscat files using dask for parallelization

Methods in this file set up a dask pipeline using futures. 
The actual logic of the map reduce is in the `map_reduce.py` file.
"""


import hipscat.io.write_metadata as io
import numpy as np
from hipscat import pixel_math
from hipscat.catalog import PartitionInfo
from hipscat.io import paths
from hipscat.io.parquet_metadata import write_parquet_metadata
from tqdm import tqdm

import hipscat_import.catalog.map_reduce as mr
from hipscat_import.catalog.arguments import ImportArguments
from hipscat_import.pipeline_resume_plan import PipelineResumePlan


def _map_pixels(args, client):
    """Generate a raw histogram of object counts in each healpix pixel"""

    if args.resume_plan.is_mapping_done():
        return

    reader_future = client.scatter(args.file_reader)
    futures = []
    for key, file_path in args.resume_plan.map_files:
        futures.append(
            client.submit(
                mr.map_to_pixels,
                input_file=file_path,
                resume_path=args.resume_plan.tmp_path,
                file_reader=reader_future,
                mapping_key=key,
                highest_order=args.mapping_healpix_order,
                ra_column=args.ra_column,
                dec_column=args.dec_column,
                use_hipscat_index=args.use_hipscat_index,
            )
        )
    args.resume_plan.wait_for_mapping(futures)


def _split_pixels(args, alignment_future, client):
    """Generate a raw histogram of object counts in each healpix pixel"""

    if args.resume_plan.is_splitting_done():
        return

    reader_future = client.scatter(args.file_reader)
    futures = []
    for key, file_path in args.resume_plan.split_keys:
        futures.append(
            client.submit(
                mr.split_pixels,
                input_file=file_path,
                file_reader=reader_future,
                highest_order=args.mapping_healpix_order,
                ra_column=args.ra_column,
                dec_column=args.dec_column,
                splitting_key=key,
                cache_shard_path=args.tmp_path,
                resume_path=args.resume_plan.tmp_path,
                alignment=alignment_future,
                use_hipscat_index=args.use_hipscat_index,
            )
        )

    args.resume_plan.wait_for_splitting(futures)


def _reduce_pixels(args, destination_pixel_map, client):
    """Loop over destination pixels and merge into parquet files"""

    if args.resume_plan.is_reducing_done():
        return

    futures = []
    for (
        destination_pixel,
        source_pixels,
        destination_pixel_key,
    ) in args.resume_plan.get_reduce_items(destination_pixel_map):
        futures.append(
            client.submit(
                mr.reduce_pixel_shards,
                cache_shard_path=args.tmp_path,
                resume_path=args.resume_plan.tmp_path,
                reducing_key=destination_pixel_key,
                destination_pixel_order=destination_pixel.order,
                destination_pixel_number=destination_pixel.pixel,
                destination_pixel_size=source_pixels[0],
                output_path=args.catalog_path,
                ra_column=args.ra_column,
                dec_column=args.dec_column,
                sort_columns=args.sort_columns,
                add_hipscat_index=args.add_hipscat_index,
                use_schema_file=args.use_schema_file,
                use_hipscat_index=args.use_hipscat_index,
                storage_options=args.output_storage_options,
            )
        )

    args.resume_plan.wait_for_reducing(futures)


def run(args, client):
    """Run catalog creation pipeline."""
    if not args:
        raise ValueError("args is required and should be type ImportArguments")
    if not isinstance(args, ImportArguments):
        raise ValueError("args must be type ImportArguments")
    _map_pixels(args, client)

    with tqdm(
        total=2, desc=PipelineResumePlan.get_formatted_stage_name("Binning"), disable=not args.progress_bar
    ) as step_progress:
        raw_histogram = args.resume_plan.read_histogram(args.mapping_healpix_order)
        step_progress.update(1)
        if args.constant_healpix_order >= 0:
            alignment = np.full(len(raw_histogram), None)
            for pixel_num, pixel_sum in enumerate(raw_histogram):
                alignment[pixel_num] = (
                    args.constant_healpix_order,
                    pixel_num,
                    pixel_sum,
                )

            destination_pixel_map = pixel_math.generate_constant_pixel_map(
                histogram=raw_histogram,
                constant_healpix_order=args.constant_healpix_order,
            )
        else:
            alignment = pixel_math.generate_alignment(
                raw_histogram,
                highest_order=args.highest_healpix_order,
                threshold=args.pixel_threshold,
            )
            destination_pixel_map = pixel_math.compute_pixel_map(
                raw_histogram,
                highest_order=args.highest_healpix_order,
                threshold=args.pixel_threshold,
            )
        step_progress.update(1)

    if not args.debug_stats_only:
        alignment_future = client.scatter(alignment)
        _split_pixels(args, alignment_future, client)
        _reduce_pixels(args, destination_pixel_map, client)

    # All done - write out the metadata
    with tqdm(
        total=5, desc=PipelineResumePlan.get_formatted_stage_name("Finishing"), disable=not args.progress_bar
    ) as step_progress:
        catalog_info = args.to_catalog_info(int(raw_histogram.sum()))
        io.write_provenance_info(
            catalog_base_dir=args.catalog_path,
            dataset_info=catalog_info,
            tool_args=args.provenance_info(),
            storage_options=args.output_storage_options,
        )
        step_progress.update(1)

        io.write_catalog_info(
            catalog_base_dir=args.catalog_path,
            dataset_info=catalog_info,
            storage_options=args.output_storage_options,
        )
        step_progress.update(1)
        partition_info = PartitionInfo.from_healpix(destination_pixel_map.keys())
        partition_info_file = paths.get_partition_info_pointer(args.catalog_path)
        partition_info.write_to_file(partition_info_file, storage_options=args.output_storage_options)
        if not args.debug_stats_only:
            write_parquet_metadata(args.catalog_path, storage_options=args.output_storage_options)
        else:
            partition_info.write_to_metadata_files(
                args.catalog_path, storage_options=args.output_storage_options
            )
        step_progress.update(1)
        io.write_fits_map(args.catalog_path, raw_histogram, storage_options=args.output_storage_options)
        step_progress.update(1)
        args.resume_plan.clean_resume_files()
        step_progress.update(1)



import warnings
from dataclasses import dataclass

import healpy as hp
from hipscat.catalog import Catalog
from hipscat.catalog.margin_cache.margin_cache_catalog_info import MarginCacheCatalogInfo
from hipscat.io.validation import is_valid_catalog

from hipscat_import.runtime_arguments import RuntimeArguments


@dataclass
class MarginCacheArguments(RuntimeArguments):
    """Container for margin cache generation arguments"""

    margin_threshold: float = 5.0
    """the size of the margin cache boundary, given in arcseconds. setting the
    `margin_threshold` to be greater than the resolution of a `margin_order`
    healpixel will result in a warning, as this may lead to data loss."""
    margin_order: int = -1
    """the order of healpixels that will be used to constrain the margin data before
    doing more precise boundary checking. this value must be greater than the highest
    order of healpix partitioning in the source catalog. if `margin_order` is left
    default or set to -1, then the `margin_order` will be set dynamically to the
    highest partition order plus 1."""

    input_catalog_path: str = ""
    """the path to the hipscat-formatted input catalog."""

    def __post_init__(self):
        self._check_arguments()

    def _check_arguments(self):
        super()._check_arguments()
        if not self.input_catalog_path:
            raise ValueError("input_catalog_path is required")
        if not is_valid_catalog(self.input_catalog_path):
            raise ValueError("input_catalog_path not a valid catalog")

        self.catalog = Catalog.read_from_hipscat(self.input_catalog_path)

        highest_order = self.catalog.partition_info.get_highest_order()
        margin_pixel_k = highest_order + 1
        if self.margin_order > -1:
            if self.margin_order < margin_pixel_k:
                raise ValueError(
                    "margin_order must be of a higher order "
                    "than the highest order catalog partition pixel."
                )
        else:
            self.margin_order = margin_pixel_k

        margin_pixel_nside = hp.order2nside(self.margin_order)

        if hp.nside2resol(margin_pixel_nside, arcmin=True) * 60.0 < self.margin_threshold:
            warnings.warn(
                "Warning: margin pixels have a smaller resolution than margin_threshold; "
                "this may lead to data loss in the margin cache."
            )

    def to_catalog_info(self, total_rows) -> MarginCacheCatalogInfo:
        """Catalog-type-specific dataset info."""
        info = {
            "catalog_name": self.output_artifact_name,
            "total_rows": total_rows,
            "catalog_type": "margin",
            "primary_catalog": self.input_catalog_path,
            "margin_threshold": self.margin_threshold,
        }
        return MarginCacheCatalogInfo(**info)

    def additional_runtime_provenance_info(self) -> dict:
        return {
            "input_catalog_path": str(self.input_catalog_path),
            "margin_threshold": self.margin_threshold,
            "margin_order": self.margin_order,
        }



import healpy as hp
import numpy as np
import pyarrow.dataset as ds
from hipscat import pixel_math
from hipscat.catalog.partition_info import PartitionInfo
from hipscat.io import file_io, paths
from hipscat.pixel_math.healpix_pixel import HealpixPixel

from hipscat_import.pipeline_resume_plan import get_pixel_cache_directory


def map_pixel_shards(
    partition_file,
    margin_pairs,
    margin_threshold,
    output_path,
    margin_order,
    ra_column,
    dec_column,
):
    """Creates margin cache shards from a source partition file."""
    data = file_io.load_parquet_to_pandas(partition_file)

    data["margin_pixel"] = hp.ang2pix(
        2**margin_order,
        data[ra_column].values,
        data[dec_column].values,
        lonlat=True,
        nest=True,
    )

    constrained_data = data.merge(margin_pairs, on="margin_pixel")

    if len(constrained_data):
        constrained_data.groupby(["partition_order", "partition_pixel"]).apply(
            _to_pixel_shard,
            margin_threshold=margin_threshold,
            output_path=output_path,
            ra_column=ra_column,
            dec_column=dec_column,
        )


def _to_pixel_shard(data, margin_threshold, output_path, ra_column, dec_column):
    """Do boundary checking for the cached partition and then output remaining data."""
    order, pix = data["partition_order"].iloc[0], data["partition_pixel"].iloc[0]
    source_order, source_pix = data["Norder"].iloc[0], data["Npix"].iloc[0]

    data["margin_check"] = pixel_math.check_margin_bounds(
        data[ra_column].values, data[dec_column].values, order, pix, margin_threshold
    )

    # pylint: disable-next=singleton-comparison
    margin_data = data.loc[data["margin_check"] == True]

    if len(margin_data):
        # generate a file name for our margin shard, that uses both sets of Norder/Npix
        partition_dir = get_pixel_cache_directory(output_path, HealpixPixel(order, pix))
        shard_dir = paths.pixel_directory(partition_dir, source_order, source_pix)

        file_io.make_directory(shard_dir, exist_ok=True)

        shard_path = paths.pixel_catalog_file(partition_dir, source_order, source_pix)

        final_df = margin_data.drop(
            columns=[
                "margin_check",
                "margin_pixel",
            ]
        )

        rename_columns = {
            PartitionInfo.METADATA_ORDER_COLUMN_NAME: f"margin_{PartitionInfo.METADATA_ORDER_COLUMN_NAME}",
            PartitionInfo.METADATA_DIR_COLUMN_NAME: f"margin_{PartitionInfo.METADATA_DIR_COLUMN_NAME}",
            PartitionInfo.METADATA_PIXEL_COLUMN_NAME: f"margin_{PartitionInfo.METADATA_PIXEL_COLUMN_NAME}",
            "partition_order": PartitionInfo.METADATA_ORDER_COLUMN_NAME,
            "partition_pixel": PartitionInfo.METADATA_PIXEL_COLUMN_NAME,
        }

        final_df.rename(columns=rename_columns, inplace=True)

        dir_column = np.floor_divide(final_df[PartitionInfo.METADATA_PIXEL_COLUMN_NAME].values, 10000) * 10000

        final_df[PartitionInfo.METADATA_DIR_COLUMN_NAME] = dir_column

        final_df = final_df.astype(
            {
                PartitionInfo.METADATA_ORDER_COLUMN_NAME: np.uint8,
                PartitionInfo.METADATA_DIR_COLUMN_NAME: np.uint64,
                PartitionInfo.METADATA_PIXEL_COLUMN_NAME: np.uint64,
            }
        )

        final_df.to_parquet(shard_path)

        del data, margin_data, final_df


def reduce_margin_shards(intermediate_directory, output_path, partition_order, partition_pixel):
    """Reduce all partition pixel directories into a single file"""
    shard_dir = get_pixel_cache_directory(
        intermediate_directory, HealpixPixel(partition_order, partition_pixel)
    )
    if file_io.does_file_or_directory_exist(shard_dir):
        data = ds.dataset(shard_dir, format="parquet")
        full_df = data.to_table().to_pandas()
        margin_cache_dir = paths.pixel_directory(output_path, partition_order, partition_pixel)
        file_io.make_directory(margin_cache_dir, exist_ok=True)

        if len(full_df):
            margin_cache_file_path = paths.pixel_catalog_file(output_path, partition_order, partition_pixel)

            full_df.to_parquet(margin_cache_file_path)
            file_io.remove_directory(shard_dir)



"""Test resume file operations"""

import os
from pathlib import Path

import pytest

from hipscat_import.pipeline_resume_plan import PipelineResumePlan


def test_done_key(tmp_path):
    """Verify expected behavior of marking stage progress via done files."""
    plan = PipelineResumePlan(tmp_path=tmp_path, progress_bar=False)
    stage = "testing"
    os.makedirs(os.path.join(tmp_path, stage))

    keys = plan.read_done_keys(stage)
    assert len(keys) == 0

    PipelineResumePlan.touch_key_done_file(tmp_path, stage, "key_01")
    keys = plan.read_done_keys(stage)
    assert keys == ["key_01"]

    PipelineResumePlan.touch_key_done_file(tmp_path, stage, "key_02")
    keys = plan.read_done_keys(stage)
    assert keys == ["key_01", "key_02"]

    plan.clean_resume_files()

    keys = plan.read_done_keys(stage)
    assert len(keys) == 0


def test_done_file(tmp_path):
    """Verify expected behavior of done file"""
    plan = PipelineResumePlan(tmp_path=tmp_path, progress_bar=False)
    done_file = "action_done"

    ## done file should not exist
    assert not plan.done_file_exists(done_file)

    ## mark done and check that we can find it
    plan.touch_stage_done_file(done_file)
    assert plan.done_file_exists(done_file)

    ## mark done is idempotent
    plan.touch_stage_done_file(done_file)
    assert plan.done_file_exists(done_file)

    ## clear all the resume files and verify done file is gone.
    plan.clean_resume_files()

    assert not plan.done_file_exists(done_file)


def test_get_keys_from_results(tmp_path):
    """Test that we can create a list of completed keys via the output results files."""
    plan = PipelineResumePlan(tmp_path=tmp_path, progress_bar=False, resume=False)
    keys = PipelineResumePlan.get_keys_from_file_names(tmp_path, ".foo")
    assert len(keys) == 0

    Path(tmp_path, "file_0.foo").touch()
    keys = PipelineResumePlan.get_keys_from_file_names(tmp_path, ".foo")
    assert keys == ["file_0"]

    Path(tmp_path, "file_1.foo").touch()
    keys = PipelineResumePlan.get_keys_from_file_names(tmp_path, ".foo")
    assert keys == ["file_0", "file_1"]

    plan.clean_resume_files()
    keys = PipelineResumePlan.get_keys_from_file_names(tmp_path, ".foo")
    assert len(keys) == 0


def test_safe_to_resume(tmp_path):
    """Check that we throw errors when it's not safe to resume."""
    plan = PipelineResumePlan(tmp_path=tmp_path, progress_bar=False, resume=False)
    plan.safe_to_resume()

    ## check is idempotent - intermediate directory exists but does not
    ## contain files.
    plan.safe_to_resume()

    ## take some action and verify that we can no longer resume without
    ## explicitly setting resume=True
    done_file = "action_done"
    plan.touch_stage_done_file(done_file)
    with pytest.raises(ValueError, match="contains intermediate"):
        plan.safe_to_resume()

    ## If we mark as a resuming pipeline, we're safe to resume.
    resuming_plan = PipelineResumePlan(tmp_path=tmp_path, progress_bar=False, resume=True)
    resuming_plan.safe_to_resume()

    ## If there are no more intermediate files, we don't need to set resume.
    plan.clean_resume_files()
    plan.safe_to_resume()


@pytest.mark.dask
def test_wait_for_futures(tmp_path, dask_client):
    """Test that we can wait around for futures to complete."""
    plan = PipelineResumePlan(tmp_path=tmp_path, progress_bar=False, resume=False)

    def error_on_even(argument):
        """Silly little method used to test futures that fail under predictable conditions"""
        if argument % 2 == 0:
            raise RuntimeError("we are at odds with evens")

    ## Everything is fine if we're all odd
    futures = [dask_client.submit(error_on_even, 1)]
    plan.wait_for_futures(futures, "test")

    ## Throw an even in the mix, and we'll see some stages fail. Should cause whole stage to fail.
    futures = [dask_client.submit(error_on_even, 1), dask_client.submit(error_on_even, 2)]
    with pytest.raises(RuntimeError, match="Some test stages failed"):
        plan.wait_for_futures(futures, "test")


def test_formatted_stage_name():
    """Test that we make pretty stage names for presenting in progress bars"""
    formatted = PipelineResumePlan.get_formatted_stage_name(None)
    assert formatted == "Progress  "

    formatted = PipelineResumePlan.get_formatted_stage_name("")
    assert formatted == "Progress  "

    formatted = PipelineResumePlan.get_formatted_stage_name("stage")
    assert formatted == "Stage     "

    formatted = PipelineResumePlan.get_formatted_stage_name("very long stage name")
    assert formatted == "Very long stage name"



"""File reading generators for common file types."""

import abc
from typing import Any, Dict, Union

import pyarrow.parquet as pq
from astropy.table import Table
from hipscat.io import FilePointer, file_io

# pylint: disable=too-few-public-methods,too-many-arguments


def get_file_reader(
    file_format,
    chunksize=500_000,
    header="infer",
    schema_file=None,
    column_names=None,
    skip_column_names=None,
    type_map=None,
    separator=",",
    **kwargs,
):
    """Get a generator file reader for common file types

    Args:
        file_format (str): specifier for the file type and extension.
            Currently supported formats include:

            - `csv`, comma separated values. may also be tab- or pipe-delimited
              includes `.csv.gz` and other compressed csv files
            - `fits`, flexible image transport system. often used for astropy tables.
            - `parquet`, compressed columnar data format

        chunksize (int): number of rows to read in a single iteration.
        header (int, list of int, None, default 'infer'): for CSV files, rows to
            use as the header with column names
        schema_file (str): path to a parquet schema file. if provided, header names
            and column types will be pulled from the parquet schema metadata.
        column_names (list[str]): for CSV files, the names of columns if no header
            is available. for fits files, a list of columns to *keep*.
        skip_column_names (list[str]): for fits files, a list of columns to remove.
        type_map (dict): for CSV files, the data types to use for columns
        separator (str): for CSV files, the character used for separation.
    """
    if "csv" in file_format:
        return CsvReader(
            chunksize=chunksize,
            header=header,
            schema_file=schema_file,
            column_names=column_names,
            type_map=type_map,
            separator=separator,
            **kwargs,
        )
    if file_format == "fits":
        return FitsReader(
            chunksize=chunksize,
            column_names=column_names,
            skip_column_names=skip_column_names,
            **kwargs,
        )
    if file_format == "parquet":
        return ParquetReader(chunksize=chunksize, **kwargs)

    raise NotImplementedError(f"File Format: {file_format} not supported")


class InputReader(abc.ABC):
    """Base class for chunking file readers."""

    @abc.abstractmethod
    def read(self, input_file):
        """Read the input file, or chunk of the input file.

        Args:
            input_file(str): path to the input file.
        Yields:
            DataFrame containing chunk of file info.
        """

    @abc.abstractmethod
    def provenance_info(self) -> dict:
        """Create dictionary of parameters for provenance tracking.

        Returns:
            dictionary with all argument_name -> argument_value as key -> value pairs.
        """

    def regular_file_exists(self, input_file, storage_options: Union[Dict[Any, Any], None] = None, **_kwargs):
        """Check that the `input_file` points to a single regular file

        Raises
            FileNotFoundError: if nothing exists at path, or directory found.
        """
        if not file_io.does_file_or_directory_exist(input_file, storage_options=storage_options):
            raise FileNotFoundError(f"File not found at path: {input_file}")
        if not file_io.is_regular_file(input_file, storage_options=storage_options):
            raise FileNotFoundError(f"Directory found at path - requires regular file: {input_file}")


class CsvReader(InputReader):
    """CSV reader for the most common CSV reading arguments.

    This uses `pandas.read_csv`, and you can find more information on
    additional arguments in the pandas documentation:
    https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html

    Attributes:
        chunksize (int): number of rows to read in a single iteration.
        header (int, list of int, None, default 'infer'): rows to
            use as the header with column names
        schema_file (str): path to a parquet schema file. if provided, header names
            and column types will be pulled from the parquet schema metadata.
        column_names (list[str]): the names of columns if no header is available
        type_map (dict): the data types to use for columns
        separator (str): the character used for separation. Use '\\s+' to
            process whitespace separated files.
    """

    def __init__(
        self,
        chunksize=500_000,
        header="infer",
        schema_file=None,
        column_names=None,
        type_map=None,
        separator=",",
        **kwargs,
    ):
        self.chunksize = chunksize
        self.header = header
        self.schema_file = schema_file
        self.column_names = column_names
        self.type_map = type_map
        self.separator = separator
        self.kwargs = kwargs

    def read(self, input_file):
        self.regular_file_exists(input_file, **self.kwargs)

        if self.schema_file:
            schema_parquet = file_io.load_parquet_to_pandas(
                FilePointer(self.schema_file),
                **self.kwargs,
            )

        use_column_names = None
        if self.column_names:
            use_column_names = self.column_names
        elif not self.header and self.schema_file:
            use_column_names = schema_parquet.columns

        use_type_map = None
        if self.type_map:
            use_type_map = self.type_map
        elif self.schema_file:
            use_type_map = schema_parquet.dtypes.to_dict()

        with file_io.load_csv_to_pandas(
            FilePointer(input_file),
            chunksize=self.chunksize,
            sep=self.separator,
            header=self.header,
            names=use_column_names,
            dtype=use_type_map,
            **self.kwargs,
        ) as reader:
            for chunk in reader:
                yield chunk

    def provenance_info(self) -> dict:
        str_type_map = {}
        if self.type_map:
            str_type_map = {key: str(value) for (key, value) in self.type_map.items()}
        provenance_info = {
            "input_reader_type": "CsvReader",
            "chunksize": self.chunksize,
            "header": self.header,
            "schema_file": self.schema_file,
            "separator": self.separator,
            "column_names": self.column_names,
            "type_map": str_type_map,
        }
        return provenance_info


class FitsReader(InputReader):
    """Chunked FITS file reader.

    There are two column-level arguments for reading fits files:
    `column_names` and `skip_column_names`.

        - If neither is provided, we will read and process all columns in the fits file.
        - If `column_names` is given, we will use *only* those names, and
          `skip_column_names` will be ignored.
        - If `skip_column_names` is provided, we will remove those columns from processing stages.

    NB:
        Uses astropy table memmap to avoid reading the entire file into memory.

        See: https://docs.astropy.org/en/stable/io/fits/index.html#working-with-large-files


    Attributes:
        chunksize (int): number of rows of the file to process at once.
            For large files, this can prevent loading the entire file
            into memory at once.
        column_names (list[str]): list of column names to keep. only use
            one of `column_names` or `skip_column_names`
        skip_column_names (list[str]): list of column names to skip. only use
            one of `column_names` or `skip_column_names`
    """

    def __init__(self, chunksize=500_000, column_names=None, skip_column_names=None, **kwargs):
        self.chunksize = chunksize
        self.column_names = column_names
        self.skip_column_names = skip_column_names
        self.kwargs = kwargs

    def read(self, input_file):
        self.regular_file_exists(input_file, **self.kwargs)
        table = Table.read(input_file, memmap=True, **self.kwargs)
        if self.column_names:
            table.keep_columns(self.column_names)
        elif self.skip_column_names:
            table.remove_columns(self.skip_column_names)

        total_rows = len(table)
        read_rows = 0

        while read_rows < total_rows:
            yield table[read_rows : read_rows + self.chunksize].to_pandas()
            read_rows += self.chunksize

    def provenance_info(self) -> dict:
        provenance_info = {
            "input_reader_type": "FitsReader",
            "chunksize": self.chunksize,
            "column_names": self.column_names,
            "skip_column_names": self.skip_column_names,
        }
        return provenance_info


class ParquetReader(InputReader):
    """Parquet reader for the most common Parquet reading arguments.

    Attributes:
        chunksize (int): number of rows of the file to process at once.
            For large files, this can prevent loading the entire file
            into memory at once.
    """

    def __init__(self, chunksize=500_000, **kwargs):
        self.chunksize = chunksize
        self.kwargs = kwargs

    def read(self, input_file):
        self.regular_file_exists(input_file, **self.kwargs)
        parquet_file = pq.ParquetFile(input_file, **self.kwargs)
        for smaller_table in parquet_file.iter_batches(batch_size=self.chunksize, use_pandas_metadata=True):
            yield smaller_table.to_pandas()

    def provenance_info(self) -> dict:
        provenance_info = {
            "input_reader_type": "ParquetReader",
            "chunksize": self.chunksize,
        }
        return provenance_info



"""Import a set of non-hipscat files using dask for parallelization"""

from typing import Any, Dict, Union

import healpy as hp
import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq
from hipscat import pixel_math
from hipscat.io import FilePointer, file_io, paths
from hipscat.pixel_math.healpix_pixel import HealpixPixel
from hipscat.pixel_math.hipscat_id import HIPSCAT_ID_COLUMN, hipscat_id_to_healpix

from hipscat_import.catalog.file_readers import InputReader
from hipscat_import.catalog.resume_plan import ResumePlan
from hipscat_import.pipeline_resume_plan import get_pixel_cache_directory

# pylint: disable=too-many-locals,too-many-arguments


def _has_named_index(dataframe):
    """Heuristic to determine if a dataframe has some meaningful index.

    This will reject dataframes with no index name for a single index,
    or empty names for multi-index (e.g. [] or [None]).
    """
    if dataframe.index.name is not None:
        ## Single index with a given name.
        return True
    if len(dataframe.index.names) == 0 or all(name is None for name in dataframe.index.names):
        return False
    return True


def _iterate_input_file(
    input_file: FilePointer,
    file_reader: InputReader,
    highest_order,
    ra_column,
    dec_column,
    use_hipscat_index=False,
):
    """Helper function to handle input file reading and healpix pixel calculation"""
    if not file_reader:
        raise NotImplementedError("No file reader implemented")

    required_columns = [ra_column, dec_column]

    for chunk_number, data in enumerate(file_reader.read(input_file)):
        if use_hipscat_index:
            if data.index.name == HIPSCAT_ID_COLUMN:
                mapped_pixels = hipscat_id_to_healpix(data.index, target_order=highest_order)
            elif HIPSCAT_ID_COLUMN in data.columns:
                mapped_pixels = hipscat_id_to_healpix(data[HIPSCAT_ID_COLUMN], target_order=highest_order)
            else:
                raise ValueError(
                    f"Invalid column names in input file: {HIPSCAT_ID_COLUMN} not in {input_file}"
                )
        else:
            if not all(x in data.columns for x in required_columns):
                raise ValueError(
                    f"Invalid column names in input file: {', '.join(required_columns)} not in {input_file}"
                )
            # Set up the pixel data
            mapped_pixels = hp.ang2pix(
                2**highest_order,
                data[ra_column].values,
                data[dec_column].values,
                lonlat=True,
                nest=True,
            )
        yield chunk_number, data, mapped_pixels


def map_to_pixels(
    input_file: FilePointer,
    file_reader: InputReader,
    resume_path: FilePointer,
    mapping_key,
    highest_order,
    ra_column,
    dec_column,
    use_hipscat_index=False,
):
    """Map a file of input objects to their healpix pixels.

    Args:
        input_file (FilePointer): file to read for catalog data.
        file_reader (hipscat_import.catalog.file_readers.InputReader): instance of input
            reader that specifies arguments necessary for reading from the input file.
        resume_path (FilePointer): where to write resume partial results.
        mapping_key (str): unique counter for this input file, used
            when creating intermediate files
        highest_order (int): healpix order to use when mapping
        ra_column (str): where to find right ascension data in the dataframe
        dec_column (str): where to find declation in the dataframe

    Returns:
        one-dimensional numpy array of long integers where the value at each index corresponds
        to the number of objects found at the healpix pixel.
    Raises:
        ValueError: if the `ra_column` or `dec_column` cannot be found in the input file.
        FileNotFoundError: if the file does not exist, or is a directory
    """
    histo = pixel_math.empty_histogram(highest_order)
    for _, _, mapped_pixels in _iterate_input_file(
        input_file, file_reader, highest_order, ra_column, dec_column, use_hipscat_index
    ):
        mapped_pixel, count_at_pixel = np.unique(mapped_pixels, return_counts=True)
        mapped_pixel = mapped_pixel.astype(np.int64)
        histo[mapped_pixel] += count_at_pixel.astype(np.int64)
    ResumePlan.write_partial_histogram(tmp_path=resume_path, mapping_key=mapping_key, histogram=histo)


def split_pixels(
    input_file: FilePointer,
    file_reader: InputReader,
    splitting_key,
    highest_order,
    ra_column,
    dec_column,
    cache_shard_path: FilePointer,
    resume_path: FilePointer,
    alignment=None,
    use_hipscat_index=False,
):
    """Map a file of input objects to their healpix pixels and split into shards.

    Args:
        input_file (FilePointer): file to read for catalog data.
        file_reader (hipscat_import.catalog.file_readers.InputReader): instance
            of input reader that specifies arguments necessary for reading from the input file.
        splitting_key (str): unique counter for this input file, used
            when creating intermediate files
        highest_order (int): healpix order to use when mapping
        ra_column (str): where to find right ascension data in the dataframe
        dec_column (str): where to find declation in the dataframe
        cache_shard_path (FilePointer): where to write intermediate parquet files.
        resume_path (FilePointer): where to write resume files.

    Raises:
        ValueError: if the `ra_column` or `dec_column` cannot be found in the input file.
        FileNotFoundError: if the file does not exist, or is a directory
    """
    for chunk_number, data, mapped_pixels in _iterate_input_file(
        input_file, file_reader, highest_order, ra_column, dec_column, use_hipscat_index
    ):
        aligned_pixels = alignment[mapped_pixels]
        unique_pixels, unique_inverse = np.unique(aligned_pixels, return_inverse=True)

        for unique_index, [order, pixel, _] in enumerate(unique_pixels):
            mapped_indexes = np.where(unique_inverse == unique_index)
            data_indexes = data.index[mapped_indexes[0].tolist()]

            filtered_data = data.filter(items=data_indexes, axis=0)

            pixel_dir = get_pixel_cache_directory(cache_shard_path, HealpixPixel(order, pixel))
            file_io.make_directory(pixel_dir, exist_ok=True)
            output_file = file_io.append_paths_to_pointer(
                pixel_dir, f"shard_{splitting_key}_{chunk_number}.parquet"
            )
            if _has_named_index(filtered_data):
                filtered_data.to_parquet(output_file, index=True)
            else:
                filtered_data.to_parquet(output_file, index=False)
        del filtered_data, data_indexes

    ResumePlan.splitting_key_done(tmp_path=resume_path, splitting_key=splitting_key)


def reduce_pixel_shards(
    cache_shard_path,
    resume_path,
    reducing_key,
    destination_pixel_order,
    destination_pixel_number,
    destination_pixel_size,
    output_path,
    ra_column,
    dec_column,
    sort_columns: str = "",
    use_hipscat_index=False,
    add_hipscat_index=True,
    delete_input_files=True,
    use_schema_file="",
    storage_options: Union[Dict[Any, Any], None] = None,
):
    """Reduce sharded source pixels into destination pixels.

    In addition to combining multiple shards of data into a single
    parquet file, this method will add a few new columns:

        - ``Norder`` - the healpix order for the pixel
        - ``Dir`` - the directory part, corresponding to the pixel
        - ``Npix`` - the healpix pixel
        - ``_hipscat_index`` - optional - a spatially-correlated
          64-bit index field.

    Notes on ``_hipscat_index``:

        - if we generate the field, we will promote any previous
          *named* pandas index field(s) to a column with
          that name.
        - see ``hipscat.pixel_math.hipscat_id``
          for more in-depth discussion of this field.

    Args:
        cache_shard_path (FilePointer): where to read intermediate parquet files.
        resume_path (FilePointer): where to write resume files.
        reducing_key (str): unique string for this task, used for resume files.
        origin_pixel_numbers (list[int]): high order pixels, with object
            data written to intermediate directories.
        destination_pixel_order (int): order of the final catalog pixel
        destination_pixel_number (int): pixel number at the above order
        destination_pixel_size (int): expected number of rows to write
            for the catalog's final pixel
        output_path (FilePointer): where to write the final catalog pixel data
        sort_columns (str): column for survey identifier, or other sortable column
        add_hipscat_index (bool): should we add a _hipscat_index column to
            the resulting parquet file?
        delete_input_files (bool): should we delete the intermediate files
            used as input for this method.
        use_schema_file (str): use the parquet schema from the indicated
            parquet file.

    Raises:
        ValueError: if the number of rows written doesn't equal provided
            `destination_pixel_size`
    """
    destination_dir = paths.pixel_directory(output_path, destination_pixel_order, destination_pixel_number)
    file_io.make_directory(destination_dir, exist_ok=True, storage_options=storage_options)

    destination_file = paths.pixel_catalog_file(
        output_path, destination_pixel_order, destination_pixel_number
    )

    schema = None
    if use_schema_file:
        schema = file_io.read_parquet_metadata(
            use_schema_file, storage_options=storage_options
        ).schema.to_arrow_schema()

    tables = []
    healpix_pixel = HealpixPixel(destination_pixel_order, destination_pixel_number)
    pixel_dir = get_pixel_cache_directory(cache_shard_path, healpix_pixel)

    if schema:
        tables.append(pq.read_table(pixel_dir, schema=schema))
    else:
        tables.append(pq.read_table(pixel_dir))

    merged_table = pa.concat_tables(tables)

    rows_written = len(merged_table)

    if rows_written != destination_pixel_size:
        raise ValueError(
            "Unexpected number of objects at pixel "
            f"({healpix_pixel})."
            f" Expected {destination_pixel_size}, wrote {rows_written}"
        )

    dataframe = merged_table.to_pandas()
    if sort_columns:
        dataframe = dataframe.sort_values(sort_columns.split(","))
    if add_hipscat_index and not use_hipscat_index:
        dataframe[HIPSCAT_ID_COLUMN] = pixel_math.compute_hipscat_id(
            dataframe[ra_column].values,
            dataframe[dec_column].values,
        )

    dataframe["Norder"] = np.full(rows_written, fill_value=healpix_pixel.order, dtype=np.uint8)
    dataframe["Dir"] = np.full(rows_written, fill_value=healpix_pixel.dir, dtype=np.uint64)
    dataframe["Npix"] = np.full(rows_written, fill_value=healpix_pixel.pixel, dtype=np.uint64)

    if add_hipscat_index:
        ## If we had a meaningful index before, preserve it as a column.
        if _has_named_index(dataframe):
            dataframe = dataframe.reset_index()
        dataframe = dataframe.set_index(HIPSCAT_ID_COLUMN).sort_index()
    dataframe.to_parquet(destination_file, storage_options=storage_options)

    del dataframe, merged_table, tables

    if delete_input_files:
        pixel_dir = get_pixel_cache_directory(cache_shard_path, healpix_pixel)

        file_io.remove_directory(pixel_dir, ignore_errors=True, storage_options=storage_options)

    ResumePlan.reducing_key_done(tmp_path=resume_path, reducing_key=reducing_key)



import pandas as pd
from dask.distributed import as_completed
from hipscat import pixel_math
from hipscat.catalog import PartitionInfo
from hipscat.io import file_io, parquet_metadata, paths, write_metadata
from tqdm import tqdm

import hipscat_import.margin_cache.margin_cache_map_reduce as mcmr

# pylint: disable=too-many-locals,too-many-arguments


def _find_partition_margin_pixel_pairs(stats, margin_order):
    """Creates a DataFrame filled with many-to-many connections between
    the catalog partition pixels and the margin pixels at `margin_order`.
    """
    norders = []
    part_pix = []
    margin_pix = []

    for healpixel in stats:
        order = healpixel.order
        pix = healpixel.pixel

        d_order = margin_order - order

        margins = pixel_math.get_margin(order, pix, d_order)

        for m_p in margins:
            norders.append(order)
            part_pix.append(pix)
            margin_pix.append(m_p)

    margin_pairs_df = pd.DataFrame(
        zip(norders, part_pix, margin_pix),
        columns=["partition_order", "partition_pixel", "margin_pixel"],
    )
    return margin_pairs_df


def _create_margin_directory(stats, output_path):
    """Creates directories for all the catalog partitions."""
    for healpixel in stats:
        order = healpixel.order
        pix = healpixel.pixel

        destination_dir = paths.pixel_directory(output_path, order, pix)
        file_io.make_directory(destination_dir, exist_ok=True)


def _map_to_margin_shards(client, args, partition_pixels, margin_pairs):
    """Create all the jobs for mapping partition files into the margin cache."""
    futures = []
    mp_future = client.scatter(margin_pairs, broadcast=True)
    for pix in partition_pixels:
        partition_file = paths.pixel_catalog_file(args.input_catalog_path, pix.order, pix.pixel)
        futures.append(
            client.submit(
                mcmr.map_pixel_shards,
                partition_file=partition_file,
                margin_pairs=mp_future,
                margin_threshold=args.margin_threshold,
                output_path=args.tmp_path,
                margin_order=args.margin_order,
                ra_column=args.catalog.catalog_info.ra_column,
                dec_column=args.catalog.catalog_info.dec_column,
            )
        )

    for _ in tqdm(
        as_completed(futures),
        desc="Mapping  ",
        total=len(futures),
        disable=not args.progress_bar,
    ):
        ...


def _reduce_margin_shards(client, args, partition_pixels):
    """Create all the jobs for reducing margin cache shards into singular files"""
    futures = []

    for pix in partition_pixels:
        futures.append(
            client.submit(
                mcmr.reduce_margin_shards,
                intermediate_directory=args.tmp_path,
                output_path=args.catalog_path,
                partition_order=pix.order,
                partition_pixel=pix.pixel,
            )
        )

    for _ in tqdm(
        as_completed(futures),
        desc="Reducing ",
        total=len(futures),
        disable=not args.progress_bar,
    ):
        ...


def generate_margin_cache(args, client):
    """Generate a margin cache for a given input catalog.
    The input catalog must be in hipscat format.

    Args:
        args (MarginCacheArguments): A valid `MarginCacheArguments` object.
        client (dask.distributed.Client): A dask distributed client object.
    """
    # determine which order to generate margin pixels for
    partition_stats = args.catalog.partition_info.get_healpix_pixels()

    # get the negative tree pixels
    negative_pixels = args.catalog.generate_negative_tree_pixels()

    combined_pixels = partition_stats + negative_pixels

    margin_pairs = _find_partition_margin_pixel_pairs(combined_pixels, args.margin_order)

    _create_margin_directory(combined_pixels, args.catalog_path)

    _map_to_margin_shards(
        client=client,
        args=args,
        partition_pixels=partition_stats,
        margin_pairs=margin_pairs,
    )

    _reduce_margin_shards(client=client, args=args, partition_pixels=combined_pixels)

    with tqdm(total=4, desc="Finishing", disable=not args.progress_bar) as step_progress:
        parquet_metadata.write_parquet_metadata(args.catalog_path)
        step_progress.update(1)
        total_rows = 0
        metadata_path = paths.get_parquet_metadata_pointer(args.catalog_path)
        for row_group in parquet_metadata.read_row_group_fragments(
            metadata_path, storage_options=args.output_storage_options
        ):
            total_rows += row_group.num_rows
        partition_info = PartitionInfo.read_from_file(
            metadata_path, storage_options=args.output_storage_options
        )
        partition_info_file = paths.get_partition_info_pointer(args.catalog_path)
        partition_info.write_to_file(partition_info_file, storage_options=args.output_storage_options)

        step_progress.update(1)
        margin_catalog_info = args.to_catalog_info(int(total_rows))
        write_metadata.write_provenance_info(
            catalog_base_dir=args.catalog_path,
            dataset_info=margin_catalog_info,
            tool_args=args.provenance_info(),
        )
        write_metadata.write_catalog_info(
            catalog_base_dir=args.catalog_path, dataset_info=margin_catalog_info
        )
        step_progress.update(1)
        file_io.remove_directory(args.tmp_path, ignore_errors=True)
        step_progress.update(1)



"""Create pixel-to-pixel association between object and source catalogs.
Methods in this file set up a dask pipeline using futures. 
The actual logic of the map reduce is in the `map_reduce.py` file.
"""

from hipscat.catalog.association_catalog.partition_join_info import PartitionJoinInfo
from hipscat.io import parquet_metadata, paths, write_metadata
from tqdm import tqdm

from hipscat_import.pipeline_resume_plan import PipelineResumePlan
from hipscat_import.soap.arguments import SoapArguments
from hipscat_import.soap.map_reduce import combine_partial_results, count_joins, reduce_joins
from hipscat_import.soap.resume_plan import SoapPlan


def run(args, client):
    """Run the association pipeline"""
    if not args:
        raise TypeError("args is required and should be type SoapArguments")
    if not isinstance(args, SoapArguments):
        raise TypeError("args must be type SoapArguments")

    resume_plan = SoapPlan(args)
    if not resume_plan.is_counting_done():
        futures = []
        for source_pixel, object_pixels, _ in resume_plan.count_keys:
            futures.append(
                client.submit(
                    count_joins,
                    soap_args=args,
                    source_pixel=source_pixel,
                    object_pixels=object_pixels,
                )
            )

        resume_plan.wait_for_counting(futures)

    if args.write_leaf_files and not resume_plan.is_reducing_done():
        for object_pixel, object_key in resume_plan.reduce_keys:
            futures.append(
                client.submit(
                    reduce_joins,
                    soap_args=args,
                    object_pixel=object_pixel,
                    object_key=object_key,
                )
            )

        resume_plan.wait_for_reducing(futures)

    # All done - write out the metadata
    with tqdm(
        total=4, desc=PipelineResumePlan.get_formatted_stage_name("Finishing"), disable=not args.progress_bar
    ) as step_progress:
        if args.write_leaf_files:
            parquet_metadata.write_parquet_metadata(args.catalog_path)
            total_rows = 0
            metadata_path = paths.get_parquet_metadata_pointer(args.catalog_path)
            for row_group in parquet_metadata.read_row_group_fragments(metadata_path):
                total_rows += row_group.num_rows
            partition_join_info = PartitionJoinInfo.read_from_file(
                metadata_path, storage_options=args.output_storage_options
            )
            partition_join_info.write_to_csv(
                catalog_path=args.catalog_path, storage_options=args.output_storage_options
            )
        else:
            total_rows = combine_partial_results(args.tmp_path, args.catalog_path)
        step_progress.update(1)
        total_rows = int(total_rows)
        catalog_info = args.to_catalog_info(total_rows)
        write_metadata.write_provenance_info(
            catalog_base_dir=args.catalog_path,
            dataset_info=catalog_info,
            tool_args=args.provenance_info(),
        )
        step_progress.update(1)
        write_metadata.write_catalog_info(dataset_info=catalog_info, catalog_base_dir=args.catalog_path)
        step_progress.update(1)
        resume_plan.clean_resume_files()
        step_progress.update(1)



import os

import pandas as pd
import pytest
from hipscat.io import paths
from hipscat.pixel_math.healpix_pixel import HealpixPixel

from hipscat_import.margin_cache import margin_cache_map_reduce
from hipscat_import.pipeline_resume_plan import get_pixel_cache_directory

keep_cols = ["weird_ra", "weird_dec"]

drop_cols = [
    "partition_order",
    "partition_pixel",
    "margin_check",
    "margin_pixel",
    "is_trunc",
]


def validate_result_dataframe(df_path, expected_len):
    res_df = pd.read_parquet(df_path)

    assert len(res_df) == expected_len

    cols = res_df.columns.values.tolist()

    for col in keep_cols:
        assert col in cols

    for col in drop_cols:
        assert col not in cols


@pytest.mark.timeout(5)
def test_to_pixel_shard_equator(tmp_path, basic_data_shard_df):
    margin_cache_map_reduce._to_pixel_shard(
        basic_data_shard_df,
        margin_threshold=360.0,
        output_path=tmp_path,
        ra_column="weird_ra",
        dec_column="weird_dec",
    )

    path = os.path.join(tmp_path, "order_1", "dir_0", "pixel_21", "Norder=1", "Dir=0", "Npix=0.parquet")

    assert os.path.exists(path)

    validate_result_dataframe(path, 2)


@pytest.mark.timeout(5)
def test_to_pixel_shard_polar(tmp_path, polar_data_shard_df):
    margin_cache_map_reduce._to_pixel_shard(
        polar_data_shard_df,
        margin_threshold=360.0,
        output_path=tmp_path,
        ra_column="weird_ra",
        dec_column="weird_dec",
    )

    path = os.path.join(tmp_path, "order_2", "dir_0", "pixel_15", "Norder=2", "Dir=0", "Npix=0.parquet")

    assert os.path.exists(path)

    validate_result_dataframe(path, 360)


@pytest.mark.dask
def test_reduce_margin_shards(tmp_path, basic_data_shard_df):
    intermediate_dir = os.path.join(tmp_path, "intermediate")
    partition_dir = get_pixel_cache_directory(intermediate_dir, HealpixPixel(1, 21))
    shard_dir = paths.pixel_directory(partition_dir, 1, 21)

    os.makedirs(shard_dir)

    first_shard_path = paths.pixel_catalog_file(partition_dir, 1, 0)
    second_shard_path = paths.pixel_catalog_file(partition_dir, 1, 1)

    shard_df = basic_data_shard_df.drop(columns=["partition_order", "partition_pixel", "margin_pixel"])

    shard_df.to_parquet(first_shard_path)
    shard_df.to_parquet(second_shard_path)

    margin_cache_map_reduce.reduce_margin_shards(intermediate_dir, tmp_path, 1, 21)

    result_path = paths.pixel_catalog_file(tmp_path, 1, 21)

    validate_result_dataframe(result_path, 720)
    assert not os.path.exists(shard_dir)



"""Utility to hold all arguments required throughout partitioning"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, List, Union

from hipscat.catalog.catalog import CatalogInfo
from hipscat.io import FilePointer
from hipscat.pixel_math import hipscat_id

from hipscat_import.catalog.file_readers import InputReader, get_file_reader
from hipscat_import.catalog.resume_plan import ResumePlan
from hipscat_import.runtime_arguments import RuntimeArguments, find_input_paths

# pylint: disable=too-many-locals,too-many-arguments,too-many-instance-attributes,too-many-branches,too-few-public-methods


@dataclass
class ImportArguments(RuntimeArguments):
    """Container class for holding partitioning arguments"""

    epoch: str = "J2000"
    """astronomical epoch for the data. defaults to "J2000" """

    catalog_type: str = "object"
    """level of catalog data, object (things in the sky) or source (detections)"""
    input_path: FilePointer | None = None
    """path to search for the input data"""
    input_format: str = ""
    """specifier of the input data format. this will be used to find an appropriate
    InputReader type, and may be used to find input files, via a match like
    ``<input_path>/*<input_format>`` """
    input_file_list: List[FilePointer] = field(default_factory=list)
    """can be used instead of `input_format` to import only specified files"""
    input_paths: List[FilePointer] = field(default_factory=list)
    """resolved list of all files that will be used in the importer"""
    input_storage_options: Union[Dict[Any, Any], None] = None
    """optional dictionary of abstract filesystem credentials for the INPUT."""

    ra_column: str = "ra"
    """column for right ascension"""
    dec_column: str = "dec"
    """column for declination"""
    use_hipscat_index: bool = False
    """use an existing hipscat spatial index as the position, instead of ra/dec"""
    sort_columns: str | None = None
    """column for survey identifier, or other sortable column. if sorting by multiple columns,
    they should be comma-separated. if `add_hipscat_index=True`, this sorting will be used to
    resolve the counter within the same higher-order pixel space"""
    add_hipscat_index: bool = True
    """add the hipscat spatial index field alongside the data"""
    use_schema_file: str | None = None
    """path to a parquet file with schema metadata. this will be used for column
    metadata when writing the files, if specified"""
    resume: bool = True
    """if there are existing intermediate resume files, should we
    read those and continue to create a new catalog where we left off"""
    constant_healpix_order: int = -1
    """healpix order to use when mapping. if this is
    a positive number, this will be the order of all final pixels and we
    will not combine pixels according to the threshold"""
    highest_healpix_order: int = 7
    """healpix order to use when mapping. this will
    not necessarily be the order used in the final catalog, as we may combine
    pixels that don't meed the threshold"""
    pixel_threshold: int = 1_000_000
    """maximum number of rows for a single resulting pixel.
    we may combine hierarchically until we near the ``pixel_threshold``"""
    mapping_healpix_order: int = -1
    """healpix order to use when mapping. will be
    ``highest_healpix_order`` unless a positive value is provided for
    ``constant_healpix_order``"""
    debug_stats_only: bool = False
    """do not perform a map reduce and don't create a new
    catalog. generate the partition info"""
    file_reader: InputReader | None = None
    """instance of input reader that specifies arguments necessary for reading
    from your input files"""
    resume_plan: ResumePlan | None = None
    """container that handles read/write of log files for this pipeline"""

    def __post_init__(self):
        self._check_arguments()

    def _check_arguments(self):
        """Check existence and consistency of argument values"""
        super()._check_arguments()

        if not self.input_format:
            raise ValueError("input_format is required")

        if self.constant_healpix_order >= 0:
            check_healpix_order_range(self.constant_healpix_order, "constant_healpix_order")
            self.mapping_healpix_order = self.constant_healpix_order
        else:
            check_healpix_order_range(self.highest_healpix_order, "highest_healpix_order")
            if not 100 <= self.pixel_threshold <= 1_000_000_000:
                raise ValueError("pixel_threshold should be between 100 and 1,000,000,000")
            self.mapping_healpix_order = self.highest_healpix_order

        if self.catalog_type not in ("source", "object"):
            raise ValueError("catalog_type should be one of `source` or `object`")

        if (not self.input_path and not self.input_file_list) or (self.input_path and self.input_file_list):
            raise ValueError("exactly one of input_path or input_file_list is required")
        if not self.file_reader:
            self.file_reader = get_file_reader(self.input_format)

        # Basic checks complete - make more checks and create directories where necessary
        self.input_paths = find_input_paths(
            self.input_path,
            f"*{self.input_format}",
            self.input_file_list,
            storage_options=self.input_storage_options,
        )
        self.resume_plan = ResumePlan(
            resume=self.resume,
            progress_bar=self.progress_bar,
            input_paths=self.input_paths,
            tmp_path=self.resume_tmp,
        )

    def to_catalog_info(self, total_rows) -> CatalogInfo:
        """Catalog-type-specific dataset info."""
        info = {
            "catalog_name": self.output_artifact_name,
            "catalog_type": self.catalog_type,
            "total_rows": total_rows,
            "epoch": self.epoch,
            "ra_column": self.ra_column,
            "dec_column": self.dec_column,
        }
        return CatalogInfo(**info)

    def additional_runtime_provenance_info(self) -> dict:
        return {
            "catalog_name": self.output_artifact_name,
            "epoch": self.epoch,
            "catalog_type": self.catalog_type,
            "input_path": str(self.input_path),
            "input_paths": self.input_paths,
            "input_format": self.input_format,
            "input_file_list": self.input_file_list,
            "ra_column": self.ra_column,
            "dec_column": self.dec_column,
            "use_hipscat_index": self.use_hipscat_index,
            "sort_columns": self.sort_columns,
            "constant_healpix_order": self.constant_healpix_order,
            "highest_healpix_order": self.highest_healpix_order,
            "pixel_threshold": self.pixel_threshold,
            "mapping_healpix_order": self.mapping_healpix_order,
            "debug_stats_only": self.debug_stats_only,
            "file_reader_info": self.file_reader.provenance_info() if self.file_reader is not None else {},
        }


def check_healpix_order_range(
    order, field_name, lower_bound=0, upper_bound=hipscat_id.HIPSCAT_ID_HEALPIX_ORDER
):
    """Helper method to check if the ``order`` is within the range determined by the
    ``lower_bound`` and ``upper_bound``, inclusive.

    Args:
        order (int): healpix order to check
        field_name (str): field name to use in the error message
        lower_bound (int): lower bound of range
        upper_bound (int): upper bound of range
    Raise:
        ValueError: if the order is outside the specified range, or bounds
            are unreasonable.
    """
    if lower_bound < 0:
        raise ValueError("healpix orders must be positive")
    if upper_bound > hipscat_id.HIPSCAT_ID_HEALPIX_ORDER:
        raise ValueError(f"healpix order should be <= {hipscat_id.HIPSCAT_ID_HEALPIX_ORDER}")
    if not lower_bound <= order <= upper_bound:
        raise ValueError(f"{field_name} should be between {lower_bound} and {upper_bound}")



"""Fixtures for testing import tool actions."""

import os
import re

import healpy as hp
import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest

# pylint: disable=missing-function-docstring, redefined-outer-name


def pytest_collection_modifyitems(config, items):
    """Modify dask unit tests to
        - ignore event loop deprecation warnings
        - have a longer timeout default timeout (5 seconds instead of 1 second)
        - require use of the `dask_client` fixture, even if it's not requsted

    Individual tests that will be particularly long-running can still override
    the default timeout, by using an annotation like:

        @pytest.mark.dask(timeout=10)
        def test_long_running():
            ...
    """
    use_ray = config.getoption("--use_ray")
    skip_ray = pytest.mark.skip(reason="skipping this test under dask-on-ray")
    first_dask = True
    for item in items:
        timeout = None
        for mark in item.iter_markers(name="dask"):
            timeout = 5
            if "timeout" in mark.kwargs:
                timeout = int(mark.kwargs.get("timeout"))
            if "skip_ray" in mark.kwargs and use_ray:
                item.add_marker(skip_ray)
        if timeout:
            if first_dask:
                ## The first test requires more time to set up the dask/ray client
                timeout += 10
                first_dask = False
            item.add_marker(pytest.mark.timeout(timeout))
            item.add_marker(pytest.mark.usefixtures("dask_client"))
            item.add_marker(pytest.mark.filterwarnings("ignore::DeprecationWarning"))


TEST_DIR = os.path.dirname(__file__)


@pytest.fixture
def test_data_dir():
    return os.path.join(TEST_DIR, "data")


@pytest.fixture
def small_sky_dir(test_data_dir):
    return os.path.join(test_data_dir, "small_sky")


@pytest.fixture
def small_sky_single_file(test_data_dir):
    return os.path.join(test_data_dir, "small_sky", "catalog.csv")


@pytest.fixture
def small_sky_object_catalog(test_data_dir):
    return os.path.join(test_data_dir, "small_sky_object_catalog")


@pytest.fixture
def small_sky_source_dir(test_data_dir):
    return os.path.join(test_data_dir, "small_sky_source")


@pytest.fixture
def small_sky_source_catalog(test_data_dir):
    return os.path.join(test_data_dir, "small_sky_source_catalog")


@pytest.fixture
def blank_data_dir(test_data_dir):
    return os.path.join(test_data_dir, "blank")


@pytest.fixture
def blank_data_file(test_data_dir):
    return os.path.join(test_data_dir, "blank", "blank.csv")


@pytest.fixture
def empty_data_dir(test_data_dir):
    return os.path.join(test_data_dir, "empty")


@pytest.fixture
def formats_dir(test_data_dir):
    return os.path.join(test_data_dir, "test_formats")


@pytest.fixture
def formats_headers_csv(test_data_dir):
    return os.path.join(test_data_dir, "test_formats", "headers.csv")


@pytest.fixture
def formats_pipe_csv(test_data_dir):
    return os.path.join(test_data_dir, "test_formats", "pipe_delimited.csv")


@pytest.fixture
def formats_fits(test_data_dir):
    return os.path.join(test_data_dir, "test_formats", "small_sky.fits")


@pytest.fixture
def formats_pandasindex(test_data_dir):
    return os.path.join(test_data_dir, "test_formats", "pandasindex.parquet")


@pytest.fixture
def formats_multiindex(test_data_dir):
    return os.path.join(test_data_dir, "test_formats", "multiindex.parquet")


@pytest.fixture
def small_sky_parts_dir(test_data_dir):
    return os.path.join(test_data_dir, "small_sky_parts")


@pytest.fixture
def small_sky_file0(test_data_dir):
    return os.path.join(test_data_dir, "small_sky_parts", "catalog_00_of_05.csv")


@pytest.fixture
def parquet_shards_dir(test_data_dir):
    return os.path.join(test_data_dir, "parquet_shards")


@pytest.fixture
def soap_intermediate_dir(test_data_dir):
    return os.path.join(test_data_dir, "soap_intermediate")


@pytest.fixture
def parquet_shards_shard_44_0(test_data_dir):
    return os.path.join(
        test_data_dir,
        "parquet_shards",
        "order_1",
        "dir_0",
        "pixel_44",
        "shard_3_0.parquet",
    )


@pytest.fixture
def mixed_schema_csv_dir(test_data_dir):
    return os.path.join(test_data_dir, "mixed_schema")


@pytest.fixture
def mixed_schema_csv_parquet(test_data_dir):
    return os.path.join(test_data_dir, "mixed_schema", "schema.parquet")


@pytest.fixture
def resume_dir(test_data_dir):
    return os.path.join(test_data_dir, "resume")


@pytest.fixture
def basic_data_shard_df():
    ras = np.arange(0.0, 360.0)
    dec = np.full(360, 0.0)
    ppix = np.full(360, 21)
    porder = np.full(360, 1)
    norder = np.full(360, 1)
    npix = np.full(360, 0)

    test_df = pd.DataFrame(
        data=zip(ras, dec, ppix, porder, norder, npix),
        columns=[
            "weird_ra",
            "weird_dec",
            "partition_pixel",
            "partition_order",
            "Norder",
            "Npix",
        ],
    )

    test_df["margin_pixel"] = hp.ang2pix(
        2**3,
        test_df["weird_ra"].values,
        test_df["weird_dec"].values,
        lonlat=True,
        nest=True,
    )

    return test_df


@pytest.fixture
def polar_data_shard_df():
    ras = np.arange(0.0, 360.0)
    dec = np.full(360, 89.9)
    ppix = np.full(360, 15)
    porder = np.full(360, 2)
    norder = np.full(360, 2)
    npix = np.full(360, 0)

    test_df = pd.DataFrame(
        data=zip(ras, dec, ppix, porder, norder, npix),
        columns=[
            "weird_ra",
            "weird_dec",
            "partition_pixel",
            "partition_order",
            "Norder",
            "Npix",
        ],
    )

    test_df["margin_pixel"] = hp.ang2pix(
        2**3,
        test_df["weird_ra"].values,
        test_df["weird_dec"].values,
        lonlat=True,
        nest=True,
    )

    return test_df


@pytest.fixture
def assert_text_file_matches():
    def assert_text_file_matches(expected_lines, file_name):
        """Convenience method to read a text file and compare the contents, line for line.

        When file contents get even a little bit big, it can be difficult to see
        the difference between an actual file and the expected contents without
        increased testing verbosity. This helper compares files line-by-line,
        using the provided strings or regular expressions.

        Notes:
            Because we check strings as regular expressions, you may need to escape some
            contents of `expected_lines`.

        Args:
            expected_lines(:obj:`string array`) list of strings, formatted as regular expressions.
            file_name (str): fully-specified path of the file to read
        """
        assert os.path.exists(file_name), f"file not found [{file_name}]"
        with open(
            file_name,
            "r",
            encoding="utf-8",
        ) as metadata_file:
            contents = metadata_file.readlines()

        assert len(expected_lines) == len(
            contents
        ), f"files not the same length ({len(contents)} vs {len(expected_lines)})"
        for i, expected in enumerate(expected_lines):
            assert re.match(expected, contents[i]), (
                f"files do not match at line {i+1} " f"(actual: [{contents[i]}] vs expected: [{expected}])"
            )

    return assert_text_file_matches


@pytest.fixture
def assert_parquet_file_ids():
    def assert_parquet_file_ids(file_name, id_column, expected_ids, resort_ids=True):
        """
        Convenience method to read a parquet file and compare the object IDs to
        a list of expected objects.

        Args:
            file_name (str): fully-specified path of the file to read
            id_column (str): column in the parquet file to read IDs from
            expected_ids (:obj:`int[]`): list of expected ids in `id_column`
        """
        assert os.path.exists(file_name), f"file not found [{file_name}]"

        data_frame = pd.read_parquet(file_name, engine="pyarrow")
        assert id_column in data_frame.columns
        ids = data_frame[id_column].tolist()
        if resort_ids:
            ids.sort()
            expected_ids.sort()

        assert len(ids) == len(
            expected_ids
        ), f"object list not the same size ({len(ids)} vs {len(expected_ids)})"

        npt.assert_array_equal(ids, expected_ids)

    return assert_parquet_file_ids


@pytest.fixture
def assert_parquet_file_index():
    def assert_parquet_file_index(file_name, expected_values):
        """
        Convenience method to read a parquet file and compare the index values to
        a list of expected objects.

        Args:
            file_name (str): fully-specified path of the file to read
            expected_values (:obj:`int[]`): list of expected values in index
        """
        assert os.path.exists(file_name), f"file not found [{file_name}]"

        data_frame = pd.read_parquet(file_name, engine="pyarrow")
        values = data_frame.index.values.tolist()
        values.sort()
        expected_values.sort()

        assert len(values) == len(
            expected_values
        ), f"object list not the same size ({len(values)} vs {len(expected_values)})"

        npt.assert_array_equal(values, expected_values)

    return assert_parquet_file_index



"""Tests of margin cache generation arguments"""


import pytest

from hipscat_import.margin_cache.margin_cache_arguments import MarginCacheArguments

# pylint: disable=protected-access


def test_empty_required(tmp_path):
    """*Most* required arguments are provided."""
    ## Input catalog path is missing
    with pytest.raises(ValueError, match="input_catalog_path"):
        MarginCacheArguments(
            margin_threshold=5.0,
            output_path=tmp_path,
            output_artifact_name="catalog_cache",
        )

    ## Input catalog path is bad
    with pytest.raises(ValueError, match="input_catalog_path"):
        MarginCacheArguments(
            input_catalog_path="/foo",
            margin_threshold=5.0,
            output_path=tmp_path,
            output_artifact_name="catalog_cache",
            overwrite=True,
        )


def test_margin_order_dynamic(small_sky_source_catalog, tmp_path):
    """Ensure we can dynamically set the margin_order"""
    args = MarginCacheArguments(
        margin_threshold=5.0,
        input_catalog_path=small_sky_source_catalog,
        output_path=tmp_path,
        output_artifact_name="catalog_cache",
    )

    assert args.margin_order == 3


def test_margin_order_static(small_sky_source_catalog, tmp_path):
    """Ensure we can manually set the margin_order"""
    args = MarginCacheArguments(
        margin_threshold=5.0,
        input_catalog_path=small_sky_source_catalog,
        output_path=tmp_path,
        output_artifact_name="catalog_cache",
        margin_order=4,
    )

    assert args.margin_order == 4


def test_margin_order_invalid(small_sky_source_catalog, tmp_path):
    """Ensure we raise an exception when margin_order is invalid"""
    with pytest.raises(ValueError, match="margin_order"):
        MarginCacheArguments(
            margin_threshold=5.0,
            input_catalog_path=small_sky_source_catalog,
            output_path=tmp_path,
            output_artifact_name="catalog_cache",
            margin_order=2,
        )


def test_margin_threshold_warns(small_sky_source_catalog, tmp_path):
    """Ensure we give a warning when margin_threshold is greater than margin_order resolution"""

    with pytest.warns(UserWarning, match="margin_threshold"):
        MarginCacheArguments(
            margin_threshold=360.0,
            input_catalog_path=small_sky_source_catalog,
            output_path=tmp_path,
            output_artifact_name="catalog_cache",
            margin_order=16,
        )


def test_to_catalog_info(small_sky_source_catalog, tmp_path):
    """Verify creation of catalog info for margin cache to be created."""
    args = MarginCacheArguments(
        margin_threshold=5.0,
        input_catalog_path=small_sky_source_catalog,
        output_path=tmp_path,
        output_artifact_name="catalog_cache",
        margin_order=4,
    )
    catalog_info = args.to_catalog_info(total_rows=10)
    assert catalog_info.catalog_name == args.output_artifact_name
    assert catalog_info.total_rows == 10


def test_provenance_info(small_sky_source_catalog, tmp_path):
    """Verify that provenance info includes margin-cache-specific fields."""
    args = MarginCacheArguments(
        margin_threshold=5.0,
        input_catalog_path=small_sky_source_catalog,
        output_path=tmp_path,
        output_artifact_name="catalog_cache",
        margin_order=4,
    )

    runtime_args = args.provenance_info()["runtime_args"]
    assert "margin_threshold" in runtime_args



"""Run pass/fail checks and generate verification report of existing hipscat table."""

from hipscat_import.verification.arguments import VerificationArguments


def run(args):
    """Run verification pipeline."""
    if not args:
        raise TypeError("args is required and should be type VerificationArguments")
    if not isinstance(args, VerificationArguments):
        raise TypeError("args must be type VerificationArguments")

    # implement everything else.
    raise NotImplementedError("Verification not yet implemented.")
